package com.sun.corba.se.impl.io;

import org.omg.CORBA.TCKind;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import org.omg.CORBA.portable.IndirectionException;
import java.lang.reflect.Array;
import com.sun.corba.se.impl.util.Utility;
import javax.rmi.CORBA.Util;
import java.rmi.Remote;
import java.io.IOException;
import com.sun.corba.se.impl.util.RepositoryId;
import com.sun.org.omg.SendingContext.CodeBaseHelper;
import org.omg.SendingContext.RunTime;
import org.omg.CORBA.portable.InputStream;
import org.omg.CORBA.portable.ValueOutputStream;
import java.io.Serializable;
import org.omg.CORBA.portable.OutputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import com.sun.corba.se.impl.logging.UtilSystemException;
import com.sun.corba.se.impl.logging.OMGSystemException;
import com.sun.org.omg.SendingContext.CodeBase;
import java.util.Hashtable;
import javax.rmi.CORBA.ValueHandlerMultiFormat;

public final class ValueHandlerImpl implements ValueHandlerMultiFormat
{
    public static final String FORMAT_VERSION_PROPERTY = "com.sun.CORBA.MaxStreamFormatVersion";
    private static final byte MAX_SUPPORTED_FORMAT_VERSION = 2;
    private static final byte STREAM_FORMAT_VERSION_1 = 1;
    private static final byte MAX_STREAM_FORMAT_VERSION;
    public static final short kRemoteType = 0;
    public static final short kAbstractType = 1;
    public static final short kValueType = 2;
    private Hashtable inputStreamPairs;
    private Hashtable outputStreamPairs;
    private CodeBase codeBase;
    private boolean useHashtables;
    private boolean isInputStream;
    private IIOPOutputStream outputStreamBridge;
    private IIOPInputStream inputStreamBridge;
    private OMGSystemException omgWrapper;
    private UtilSystemException utilWrapper;
    
    private static byte getMaxStreamFormatVersion() {
        try {
            String s = (String)AccessController.doPrivileged(new PrivilegedAction() {
                @Override
                public Object run() {
                    return System.getProperty("com.sun.CORBA.MaxStreamFormatVersion");
                }
            });
            if (s == null) {
                return 2;
            }
            int byte1 = Byte.parseByte(s);
            if (byte1 < 1 || byte1 > 2) {
                throw new ExceptionInInitializerError("Invalid stream format version: " + byte1 + ".  Valid range is 1 through " + 2);
            }
            return byte1;
        }
        catch (Exception ex) {
            ExceptionInInitializerError exceptionInInitializerError = new ExceptionInInitializerError(ex);
            exceptionInInitializerError.initCause(ex);
            throw exceptionInInitializerError;
        }
    }
    
    public byte getMaximumStreamFormatVersion() {
        return ValueHandlerImpl.MAX_STREAM_FORMAT_VERSION;
    }
    
    public void writeValue(OutputStream outputStream, Serializable s, byte value) {
        if (value == 2) {
            if (!(outputStream instanceof ValueOutputStream)) {
                throw this.omgWrapper.notAValueoutputstream();
            }
        }
        else if (value != 1) {
            throw this.omgWrapper.invalidStreamFormatVersion(new Integer((int)value));
        }
        this.writeValueWithVersion(outputStream, s, value);
    }
    
    private ValueHandlerImpl() {
        this.inputStreamPairs = null;
        this.outputStreamPairs = null;
        this.codeBase = null;
        this.useHashtables = true;
        this.isInputStream = true;
        this.outputStreamBridge = null;
        this.inputStreamBridge = null;
        this.omgWrapper = OMGSystemException.get("rpc.encoding");
        this.utilWrapper = UtilSystemException.get("rpc.encoding");
    }
    
    private ValueHandlerImpl(boolean isInputStream) {
        this();
        this.useHashtables = false;
        this.isInputStream = isInputStream;
    }
    
    static ValueHandlerImpl getInstance() {
        return new ValueHandlerImpl();
    }
    
    static ValueHandlerImpl getInstance(boolean b) {
        return new ValueHandlerImpl(b);
    }
    
    public void writeValue(OutputStream outputStream, Serializable s) {
        this.writeValueWithVersion(outputStream, s, (byte)1);
    }
    
    private void writeValueWithVersion(OutputStream key, Serializable s, byte b) {
        org.omg.CORBA_2_3.portable.OutputStream outputStream = (org.omg.CORBA_2_3.portable.OutputStream)key;
        if (!this.useHashtables) {
            if (this.outputStreamBridge == null) {
                (this.outputStreamBridge = this.createOutputStream()).setOrbStream(outputStream);
            }
            try {
                this.outputStreamBridge.increaseRecursionDepth();
                this.writeValueInternal(this.outputStreamBridge, outputStream, s, b);
            }
            finally {
                this.outputStreamBridge.decreaseRecursionDepth();
            }
            return;
        }
        else {
            if (this.outputStreamPairs == null) {
                this.outputStreamPairs = new Hashtable();
            }
            IIOPOutputStream outputStream2 = (IIOPOutputStream)this.outputStreamPairs.get(key);
            if (outputStream2 == null) {
                outputStream2 = this.createOutputStream();
                outputStream2.setOrbStream(outputStream);
                this.outputStreamPairs.put(key, outputStream2);
            }
            try {
                outputStream2.increaseRecursionDepth();
                this.writeValueInternal(outputStream2, outputStream, s, b);
            }
            finally {
                if (outputStream2.decreaseRecursionDepth() == 0) {
                    this.outputStreamPairs.remove(key);
                }
            }
            return;
        }
    }
    
    private void writeValueInternal(IIOPOutputStream iiopOutputStream, org.omg.CORBA_2_3.portable.OutputStream outputStream, Serializable s, byte b) {
        Class<Object> class1 = s.getClass();
        if (class1.isArray()) {
            this.write_Array(outputStream, s, class1.getComponentType());
        }
        else {
            iiopOutputStream.simpleWriteObject(s, b);
        }
    }
    
    public Serializable readValue(InputStream key, int n, Class clazz, String s, RunTime runTime) {
        CodeBase narrow = CodeBaseHelper.narrow(runTime);
        org.omg.CORBA_2_3.portable.InputStream inputStream = (org.omg.CORBA_2_3.portable.InputStream)key;
        if (!this.useHashtables) {
            if (this.inputStreamBridge == null) {
                (this.inputStreamBridge = this.createInputStream()).setOrbStream(inputStream);
                this.inputStreamBridge.setSender(narrow);
                this.inputStreamBridge.setValueHandler(this);
            }
            Serializable valueInternal2 = null;
            try {
                this.inputStreamBridge.increaseRecursionDepth();
                valueInternal2 = this.readValueInternal(this.inputStreamBridge, inputStream, n, clazz, s, narrow);
            }
            finally {
                if (this.inputStreamBridge.decreaseRecursionDepth() != 0) {}
            }
            return valueInternal2;
        }
        else {
            if (this.inputStreamPairs == null) {
                this.inputStreamPairs = new Hashtable();
            }
            IIOPInputStream inputStream2 = (IIOPInputStream)this.inputStreamPairs.get(key);
            if (inputStream2 == null) {
                inputStream2 = this.createInputStream();
                inputStream2.setOrbStream(inputStream);
                inputStream2.setSender(narrow);
                inputStream2.setValueHandler(this);
                this.inputStreamPairs.put(key, inputStream2);
            }
            Serializable valueInternal = null;
            try {
                inputStream2.increaseRecursionDepth();
                valueInternal = this.readValueInternal(inputStream2, inputStream, n, clazz, s, narrow);
            }
            finally {
                if (inputStream2.decreaseRecursionDepth() == 0) {
                    this.inputStreamPairs.remove(key);
                }
            }
            return valueInternal;
        }
    }
    
    private Serializable readValueInternal(IIOPInputStream iiopInputStream, org.omg.CORBA_2_3.portable.InputStream inputStream, int n, Class clazz, String s, CodeBase codeBase) {
        __Null _Null = null;
        if (clazz == null) {
            if (this.isArray(s)) {
                this.read_Array(iiopInputStream, inputStream, null, codeBase, n);
                return _Null;
            }
            iiopInputStream.simpleSkipObject(s, codeBase);
            return _Null;
        }
        else {
            if (clazz.isArray()) {
                return (Serializable)this.read_Array(iiopInputStream, inputStream, clazz, codeBase, n);
            }
            return (Serializable)iiopInputStream.simpleReadObject(clazz, s, codeBase, n);
        }
    }
    
    public String getRMIRepositoryID(Class clazz) {
        return RepositoryId.createForJavaType(clazz);
    }
    
    public boolean isCustomMarshaled(Class clazz) {
        return ObjectStreamClass.lookup(clazz).isCustomMarshaled();
    }
    
    public RunTime getRunTimeCodeBase() {
        if (this.codeBase != null) {
            return this.codeBase;
        }
        this.codeBase = new FVDCodeBaseImpl();
        ((FVDCodeBaseImpl)this.codeBase).setValueHandler(this);
        return this.codeBase;
    }
    
    public boolean useFullValueDescription(Class clazz, String s) throws IOException {
        return RepositoryId.useFullValueDescription(clazz, s);
    }
    
    public String getClassName(String s) {
        return RepositoryId.cache.getId(s).getClassName();
    }
    
    public Class getClassFromType(String s) throws ClassNotFoundException {
        return RepositoryId.cache.getId(s).getClassFromType();
    }
    
    public Class getAnyClassFromType(String s) throws ClassNotFoundException {
        return RepositoryId.cache.getId(s).getAnyClassFromType();
    }
    
    public String createForAnyType(Class clazz) {
        return RepositoryId.createForAnyType(clazz);
    }
    
    public String getDefinedInId(String s) {
        return RepositoryId.cache.getId(s).getDefinedInId();
    }
    
    public String getUnqualifiedName(String s) {
        return RepositoryId.cache.getId(s).getUnqualifiedName();
    }
    
    public String getSerialVersionUID(String s) {
        return RepositoryId.cache.getId(s).getSerialVersionUID();
    }
    
    public boolean isAbstractBase(Class clazz) {
        return RepositoryId.isAbstractBase(clazz);
    }
    
    public boolean isSequence(String s) {
        return RepositoryId.cache.getId(s).isSequence();
    }
    
    public Serializable writeReplace(Serializable s) {
        return ObjectStreamClass.lookup(s.getClass()).writeReplace(s);
    }
    
    private void writeCharArray(org.omg.CORBA_2_3.portable.OutputStream outputStream, char[] array, int n, int n2) {
        outputStream.write_wchar_array(array, n, n2);
    }
    
    private void write_Array(org.omg.CORBA_2_3.portable.OutputStream outputStream, Serializable s, Class clazz) {
        if (clazz.isPrimitive()) {
            if (clazz == Integer.TYPE) {
                int[] array10 = (int[])s;
                int length10 = array10.length;
                outputStream.write_ulong(length10);
                outputStream.write_long_array(array10, 0, length10);
            }
            else if (clazz == Byte.TYPE) {
                byte[] array9 = (byte[])s;
                int length9 = array9.length;
                outputStream.write_ulong(length9);
                outputStream.write_octet_array(array9, 0, length9);
            }
            else if (clazz == Long.TYPE) {
                long[] array8 = (long[])s;
                int length8 = array8.length;
                outputStream.write_ulong(length8);
                outputStream.write_longlong_array(array8, 0, length8);
            }
            else if (clazz == Float.TYPE) {
                float[] array7 = (float[])s;
                int length7 = array7.length;
                outputStream.write_ulong(length7);
                outputStream.write_float_array(array7, 0, length7);
            }
            else if (clazz == Double.TYPE) {
                double[] array6 = (double[])s;
                int length6 = array6.length;
                outputStream.write_ulong(length6);
                outputStream.write_double_array(array6, 0, length6);
            }
            else if (clazz == Short.TYPE) {
                short[] array5 = (short[])s;
                int length5 = array5.length;
                outputStream.write_ulong(length5);
                outputStream.write_short_array(array5, 0, length5);
            }
            else if (clazz == Character.TYPE) {
                char[] array4 = (char[])s;
                int length4 = array4.length;
                outputStream.write_ulong(length4);
                this.writeCharArray(outputStream, array4, 0, length4);
            }
            else if (clazz == Boolean.TYPE) {
                boolean[] array3 = (boolean[])s;
                int length3 = array3.length;
                outputStream.write_ulong(length3);
                outputStream.write_boolean_array(array3, 0, length3);
            }
            else {
                throw new Error("Invalid primitive type : " + s.getClass().getName());
            }
        }
        else if (clazz == Object.class) {
            Object[] array2 = (Object[])s;
            int length2 = array2.length;
            outputStream.write_ulong(length2);
            for (int j = 0; j < length2; ++j) {
                Util.writeAny(outputStream, array2[j]);
            }
        }
        else {
            Object[] array = (Object[])s;
            int length = array.length;
            outputStream.write_ulong(length);
            int n = 2;
            if (clazz.isInterface()) {
                clazz.getName();
                if (Remote.class.isAssignableFrom(clazz)) {
                    n = 0;
                }
                else if (org.omg.CORBA.Object.class.isAssignableFrom(clazz)) {
                    n = 0;
                }
                else if (RepositoryId.isAbstractBase(clazz)) {
                    n = 1;
                }
                else if (ObjectStreamClassCorbaExt.isAbstractInterface(clazz)) {
                    n = 1;
                }
            }
            for (int i = 0; i < length; ++i) {
                switch (n) {
                    case 0: {
                        Util.writeRemoteObject(outputStream, array[i]);
                        break;
                    }
                    case 1: {
                        Util.writeAbstractObject(outputStream, array[i]);
                        break;
                    }
                    case 2: {
                        try {
                            outputStream.write_value((Serializable)array[i]);
                            break;
                        }
                        catch (ClassCastException ex) {
                            if ((array[i] instanceof Serializable)) {
                                throw ex;
                            }
                            Utility.throwNotSerializableForCorba(array[i].getClass().getName());
                        }
                        break;
                    }
                }
            }
        }
    }
    
    private void readCharArray(org.omg.CORBA_2_3.portable.InputStream inputStream, char[] array, int n, int n2) {
        inputStream.read_wchar_array(array, n, n2);
    }
    
    private Object read_Array(IIOPInputStream iiopInputStream, org.omg.CORBA_2_3.portable.InputStream inputStream, Class clazz, CodeBase codeBase, int n) {
        try {
            int read_ulong = inputStream.read_ulong();
            if (clazz == null) {
                for (int k = 0; k < read_ulong; ++k) {
                    inputStream.read_value();
                }
                return null;
            }
            Class<Object> clazz2 = clazz.getComponentType();
            if (clazz2.isPrimitive()) {
                if (clazz2 == Integer.TYPE) {
                    int[] array10 = new int[read_ulong];
                    inputStream.read_long_array(array10, 0, read_ulong);
                    return (Serializable)array10;
                }
                if (clazz2 == Byte.TYPE) {
                    byte[] array9 = new byte[read_ulong];
                    inputStream.read_octet_array(array9, 0, read_ulong);
                    return (Serializable)array9;
                }
                if (clazz2 == Long.TYPE) {
                    long[] array8 = new long[read_ulong];
                    inputStream.read_longlong_array(array8, 0, read_ulong);
                    return (Serializable)array8;
                }
                if (clazz2 == Float.TYPE) {
                    float[] array7 = new float[read_ulong];
                    inputStream.read_float_array(array7, 0, read_ulong);
                    return (Serializable)array7;
                }
                if (clazz2 == Double.TYPE) {
                    double[] array6 = new double[read_ulong];
                    inputStream.read_double_array(array6, 0, read_ulong);
                    return (Serializable)array6;
                }
                if (clazz2 == Short.TYPE) {
                    short[] array5 = new short[read_ulong];
                    inputStream.read_short_array(array5, 0, read_ulong);
                    return (Serializable)array5;
                }
                if (clazz2 == Character.TYPE) {
                    char[] array4 = new char[read_ulong];
                    this.readCharArray(inputStream, array4, 0, read_ulong);
                    return (Serializable)array4;
                }
                if (clazz2 == Boolean.TYPE) {
                    boolean[] array3 = new boolean[read_ulong];
                    inputStream.read_boolean_array(array3, 0, read_ulong);
                    return (Serializable)array3;
                }
                throw new Error("Invalid primitive componentType : " + clazz.getName());
            }
            else {
                if (clazz2 == Object.class) {
                    Object[] array2 = (Object[])(Object[])Array.newInstance(clazz2, read_ulong);
                    iiopInputStream.activeRecursionMgr.addObject(n, array2);
                    Object o;
                    for (int j = 0; j < read_ulong; array2[j] = o, ++j) {
                        try {
                            o = Util.readAny(inputStream);
                        }
                        catch (IndirectionException ex4) {
                            try {
                                o = iiopInputStream.activeRecursionMgr.getObject(ex4.offset);
                            }
                            catch (IOException ex5) {
                                throw this.utilWrapper.invalidIndirection(ex5, new Integer(ex4.offset));
                            }
                        }
                    }
                    return (Serializable)array2;
                }
                Object[] array = (Object[])(Object[])Array.newInstance(clazz2, read_ulong);
                iiopInputStream.activeRecursionMgr.addObject(n, array);
                int n2 = 2;
                boolean n3 = false;
                if (clazz2.isInterface()) {
                    boolean n4 = false;
                    if (Remote.class.isAssignableFrom(clazz2)) {
                        n2 = 0;
                        n4 = true;
                    }
                    else if (org.omg.CORBA.Object.class.isAssignableFrom(clazz2)) {
                        n2 = 0;
                        n4 = true;
                    }
                    else if (RepositoryId.isAbstractBase(clazz2)) {
                        n2 = 1;
                        n4 = true;
                    }
                    else if (ObjectStreamClassCorbaExt.isAbstractInterface(clazz2)) {
                        n2 = 1;
                    }
                    if (!(n4 == 0)) {
                        try {
                            clazz2 = Utility.loadStubClass(RepositoryId.createForAnyType(clazz2), Util.getCodebase(clazz2), clazz2);
                        }
                        catch (ClassNotFoundException ex) {
                            n3 = true;
                        }
                    }
                    else {
                        n3 = true;
                    }
                }
                for (int i = 0; i < read_ulong; ++i) {
                    try {
                        switch (n2) {
                            case 0: {
                                if (n3 == 0) {
                                    array[i] = inputStream.read_Object(clazz2);
                                    break;
                                }
                                array[i] = Utility.readObjectAndNarrow(inputStream, clazz2);
                                break;
                            }
                            case 1: {
                                if (n3 == 0) {
                                    array[i] = inputStream.read_abstract_interface(clazz2);
                                    break;
                                }
                                array[i] = Utility.readAbstractAndNarrow(inputStream, clazz2);
                                break;
                            }
                            case 2: {
                                array[i] = inputStream.read_value(clazz2);
                                break;
                            }
                        }
                    }
                    catch (IndirectionException ex2) {
                        try {
                            array[i] = iiopInputStream.activeRecursionMgr.getObject(ex2.offset);
                        }
                        catch (IOException ex3) {
                            throw this.utilWrapper.invalidIndirection(ex3, new Integer(ex2.offset));
                        }
                    }
                }
                return (Serializable)array;
            }
        }
        finally {
            iiopInputStream.activeRecursionMgr.removeObject(n);
        }
    }
    
    private boolean isArray(String s) {
        return RepositoryId.cache.getId(s).isSequence();
    }
    
    private String getOutputStreamClassName() {
        return "com.sun.corba.se.impl.io.IIOPOutputStream";
    }
    
    private IIOPOutputStream createOutputStream() {
        String outputStreamClassName = this.getOutputStreamClassName();
        try {
            IIOPOutputStream outputStreamBuiltIn = this.createOutputStreamBuiltIn(outputStreamClassName);
            if (outputStreamBuiltIn != null) {
                return outputStreamBuiltIn;
            }
            return (IIOPOutputStream)this.createCustom(IIOPOutputStream.class, outputStreamClassName);
        }
        catch (Throwable cause) {
            InternalError internalError = new InternalError("Error loading " + outputStreamClassName);
            internalError.initCause(cause);
            throw internalError;
        }
    }
    
    private IIOPOutputStream createOutputStreamBuiltIn(final String s) throws Throwable {
        try {
            return (IIOPOutputStream)AccessController.doPrivileged(new PrivilegedExceptionAction<IIOPOutputStream>() {
                @Override
                public IIOPOutputStream run() throws IOException {
                    return ValueHandlerImpl.access$000(ValueHandlerImpl.this, s);
                }
            });
        }
        catch (PrivilegedActionException ex) {
            throw ex.getCause();
        }
    }
    
    private IIOPOutputStream createOutputStreamBuiltInNoPriv(String s) throws IOException {
        return s.equals(IIOPOutputStream.class.getName()) ? new IIOPOutputStream() : null;
    }
    
    private String getInputStreamClassName() {
        return "com.sun.corba.se.impl.io.IIOPInputStream";
    }
    
    private IIOPInputStream createInputStream() {
        String inputStreamClassName = this.getInputStreamClassName();
        try {
            IIOPInputStream inputStreamBuiltIn = this.createInputStreamBuiltIn(inputStreamClassName);
            if (inputStreamBuiltIn != null) {
                return inputStreamBuiltIn;
            }
            return (IIOPInputStream)this.createCustom(IIOPInputStream.class, inputStreamClassName);
        }
        catch (Throwable cause) {
            InternalError internalError = new InternalError("Error loading " + inputStreamClassName);
            internalError.initCause(cause);
            throw internalError;
        }
    }
    
    private IIOPInputStream createInputStreamBuiltIn(final String s) throws Throwable {
        try {
            return (IIOPInputStream)AccessController.doPrivileged(new PrivilegedExceptionAction<IIOPInputStream>() {
                @Override
                public IIOPInputStream run() throws IOException {
                    return ValueHandlerImpl.access$100(ValueHandlerImpl.this, s);
                }
            });
        }
        catch (PrivilegedActionException ex) {
            throw ex.getCause();
        }
    }
    
    private IIOPInputStream createInputStreamBuiltInNoPriv(String s) throws IOException {
        return s.equals(IIOPInputStream.class.getName()) ? new IIOPInputStream() : null;
    }
    
    private <T> T createCustom(Class<T> clazz, String name) throws Throwable {
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        if (classLoader == null) {
            classLoader = ClassLoader.getSystemClassLoader();
        }
        return classLoader.loadClass(name).asSubclass(clazz).newInstance();
    }
    
    TCKind getJavaCharTCKind() {
        return TCKind.tk_wchar;
    }
    
    static IIOPOutputStream access$000(ValueHandlerImpl valueHandlerImpl, String s) throws IOException {
        return valueHandlerImpl.createOutputStreamBuiltInNoPriv(s);
    }
    
    static IIOPInputStream access$100(ValueHandlerImpl valueHandlerImpl, String s) throws IOException {
        return valueHandlerImpl.createInputStreamBuiltInNoPriv(s);
    }
    
    static {
        MAX_STREAM_FORMAT_VERSION = getMaxStreamFormatVersion();
    }
}
